/*******************************************************************************
 * File:          riar_linearaxis_demo.ino                                      *
 * Version:       4.1.0 (Unmaintained)                                          *
 * Last Updated:  2023-05-01 (YYYY-MM-DD)                                       *
 * Author:        Görkem Can Ertemli                                            *
 * Description:   Controls the movement of a linear axis using NodeRED and      *
 *                NodeMCU, integrating with MQTT for communication.             *
********************************************************************************/

// -----------------------------------------------------------------------------
// ---------------------------- Declaration Section ----------------------------
// -----------------------------------------------------------------------------

#include <AccelStepper.h>                               // Library for stepper motor control
#include <ArduinoJson.h>                                // JSON format parsing library
#include "EspMQTTClient.h"                              // Simple WIFI & MQTT Library

// WiFi and MQTT Configuration
#define WIFI_NAME "KIA_2GHZ"                            // WiFi Name 
#define WIFI_PASSWORD "********"                        // WiFi Password 
#define MQTT_BROKER_IP "192.168.1.2"                    // MQTT Broker server IP local
#define CLIENT_NAME "linaxis1"                          // Client name that uniquely identifies your device
#define MQTT_PORT 1883                                  // The MQTT port, default to 1883. This line can be omitted

String DeviceName = "linaxis1";                         // Device identifier

EspMQTTClient client(
  WIFI_NAME,
  WIFI_PASSWORD,
  MQTT_BROKER_IP,
  CLIENT_NAME,
  MQTT_PORT
);

// External Interrupt for Limit Switch on Arduino UNO
#define SWITCH_PIN D5                                     // D5

// Connection between Node MCU and Motor Driver 
#define DIR_PIN D7                                        // D7 
#define PUL_PIN D6                                        // D6

// '1' means DRIVER option of Accelstepper is selected
AccelStepper stepper(1, PUL_PIN, DIR_PIN);

long int steps;                                            // To store the steps entered by user
long int distanceMM;                                       // To store the steps entered by user in milimeters 
long int ACCL;                                             // To store the acceleration entered by user
long int MAX_SPEED;                                        // To store the max speed of the motor entered by user

String command;                                            // To store the command sent by the user

// -----------------------------------------------------------------------------
// ------------------------------- Setup Section -------------------------------
// -----------------------------------------------------------------------------
// Setup function: initializes the system

void setup()
{
  Serial.begin(115200);                                     // Start serial communication
  client.enableDebuggingMessages();                         // Enable debugging messages for MQTT client
  
  // Set default values if not provided
  MAX_SPEED = (MAX_SPEED == 0) ? 1600 : MAX_SPEED;
  ACCL = (ACCL == 0) ? 1600 : ACCL;
  
  // Set maximum speed and acceleration for the stepper motor
  stepper.setMaxSpeed(MAX_SPEED);
  stepper.setAcceleration(ACCL);

  pinMode(SWITCH_PIN, INPUT_PULLUP);                        // Set the limit switch pin as input with internal pull-up resistor

  home();                                                   // Execute the home function before entering the loop
}

// -----------------------------------------------------------------------------
// ------------------------------- Loop Section --------------------------------
// -----------------------------------------------------------------------------
// Loop function: runs continuously after setup

void loop()
{
  client.loop();                                             // Maintain MQTT client connection                      
      
  // Publish device states to MQTT broker
  if (client.isConnected()) {
   client.publish("devices", DeviceName);
   delay(500);   
  }  
}

// -----------------------------------------------------------------------------
// ------------------------------- Functions Section -------------------------------
// -----------------------------------------------------------------------------

// Home function: homes the linear axis
void home()
{
  long int stepsToZero = 100;
  ACCL = 1600;
  MAX_SPEED = 1600;

  // Negative steps allow the motor to move in the clockwise direction.
  // The limit switch is assumed to be placed on the side of the motor.
  // -30000 is entered to make sure that the limit switch is encountered.
  stepper.moveTo(-30000);

  while (stepper.distanceToGo() != 0)
  {
    stepper.run();

    // following if condition is executed if the limit switch is encountered
    if (!digitalRead(SWITCH_PIN))
    {
      Serial.println("Reached home!!");
      Serial.println("Moving to new position");

      // The following lines move the block away from the limit switch, so that it is no longer pressed.
      stepper.setCurrentPosition(0);                            // sets new zero position for the motor      
      stepper.moveTo(stepsToZero);                              // gives new position for motor to move away from limit switch
      
    }

    yield();
  }

  // New zero position, which is away from limit switch
  stepper.setCurrentPosition(0);
}

void onConnectionEstablished(){  
  // subscribe to the device cmd input when a connection is established
  client.subscribe(DeviceName + "/cmd", onMessageReceived);    
}

void onMessageReceived(const String &payload)
{
    Serial.println(payload);

    DynamicJsonDocument doc(200);
    deserializeJson(doc, payload);                               // deserialize incoming payload

    command = (String)doc["cmd"];
    distanceMM = (int)doc["meta"]["distance"];
    ACCL = (int)doc["meta"]["accel"];
    MAX_SPEED = (int)doc["meta"]["maxspeed"];

    // Update meta values
    if (doc.containsKey("meta"))
    {
        JsonObject meta = doc["meta"].as<JsonObject>();
        if (meta.containsKey("maxspeed"))
        {
            MAX_SPEED = meta["maxspeed"].as<int>();
            stepper.setMaxSpeed(MAX_SPEED);
        }
        if (meta.containsKey("accel"))
        {
            ACCL = meta["accel"].as<int>();
            stepper.setAcceleration(ACCL);
        }
    }

    if (command == "Home")
    {
        home();
    }
    else if (command == "SetValue")
    {
        Serial.print("The distance entered by the user is: ");
        Serial.println(distanceMM);
        Serial.print("mm");

        if (distanceMM < 0)
        {
            distanceMM = 0;
        }
        else if (distanceMM >= 280)
        {
            distanceMM = 280;
        }

        steps = distanceMM * 80;

        if (digitalRead(SWITCH_PIN))
        {
            stepper.moveTo(steps);
            stepper.runToPosition();
        }
    }  
}